#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Post QA: Retrieve results directories from QA2 area to the spool area """

# pylint: disable=E0401, E0402, W1203

import glob
import logging
import os
import shutil
from pathlib import Path
from typing import Dict, List, Tuple

import requests
from requests import Response

from .interfaces import ConveyorIF


class RetrievalConveyor(ConveyorIF):
    """Retrieves results directories from QA2 area to the spool area"""

    def __init__(self, settings: Dict[str, str], action: str):
        self.logger = logging.getLogger("conveyor")
        self.action = action
        self.settings = settings

    def retrieval(self):
        """
        Get results from QA2 area and move them to spool dir

        :return:
        """
        qa_path = self.determine_qa_directory()
        spool_path = self.determine_spool_directory()

        os.chdir(spool_path)

        contents = glob.glob("*/", recursive=True)

        self.break_symlinks(spool_path, contents)
        if qa_path.exists():
            self.move_subdirectories_to_spool(qa_path, spool_path, contents)
            check = self.check_spool_contents(spool_path, contents)
            if check:
                Path(qa_path).rmdir()
        else:
            self.logger.info(f"Directories from {qa_path} have already been retrieved! Continuing...")

        self.save_qa_notes(spool_path)

        if self.action == "QA Fail Cleanup":
            self.send_qa_fail_msg(self.settings["sdm_ids"])

    def save_qa_notes(self, spool_path):
        # obtain the qa_notes.html data and push it up to the capability version here
        try:
            # the path will be something like products/pipeline-YYYYMMDDTHHMMDD/html/qa_notes.html
            qa_notes_file = next((spool_path / Path("products")).glob("pipeline-*")) / "html" / "qa_notes.html"

            # does it exist?
            if qa_notes_file.exists():
                # read the file
                qa_notes = qa_notes_file.read_text()

                # calculate the URL
                request, version = self.determine_capability_request_and_version_ids()
                url = (
                    f"{self.settings['capability_url']}"
                    f"/capability/request/{request}"
                    f"/version/{version}"
                    f"/workflow_metadata"
                )

                # push it up
                requests.post(url, json={"qa": qa_notes})
            else:
                self.logger.warning("No qa_notes.html file found; unable to persist QA notes")
        except:
            self.logger.warning("Unable to transmit QA notes to capability service")

    def send_qa_fail_msg(self, sdm_id: str) -> Response:
        """
        Call REST endpoint that shoots off QA Fail message

        :return:
        """
        url = (
            f"{self.settings['workflow_url']}/workflows/requests/"
            f"{self.settings['request_id']}/send-do-not-calibrate/{sdm_id}"
        )
        return requests.post(url, json={"sdm_id": sdm_id})

    def break_symlinks(self, spool_path: Path, dir_list: List[str]):
        """
        Lose the symlinks we needed temporarily

        :param spool_path:
        :param dir_list:
        :return:
        """
        self.logger.info(f"Breaking symlinks between qa2 and spool for directory {spool_path.stem}...")

        for directory in dir_list:
            directory = directory.strip("/")
            if Path(spool_path / directory).is_symlink():
                Path(spool_path / directory).unlink()
                self.logger.info(f'Symlink for "{directory}" removed.')
            else:
                self.logger.info(f'Directory "{directory}" is not a symlink! Skipping.')

    def move_subdirectories_to_spool(self, qa_path: Path, spool_path: Path, dir_list: List[str]):
        """
        Moves QA2 subdirectories to the request's spool dir.

        :param qa_path:
        :param spool_path:
        :param dir_list:
        :return:
        """
        self.logger.info(f"Moving directories from qa2 to spool for directory {qa_path.stem}...")
        for directory in dir_list:
            qa_dir = qa_path / directory
            spool_dir = spool_path / directory
            self.logger.info(f"Moving directory {directory}...")
            shutil.move(qa_dir, spool_dir)
            self.logger.info("Done.")

    def determine_qa_directory(self) -> Path:
        """
        Find the QA2 directory for this request.

        :return:
        """
        if "Calibration" in self.action:
            qa_area = self.settings["qa_delivery_area"]
        else:
            qa_area = self.settings["image_qa_area"]

        wf_dir = self.settings["destination_subdir"]
        return Path(qa_area + "/" + wf_dir)

    def determine_spool_directory(self) -> Path:
        """
        Find the spool directory for this request.

        :return:
        """
        spool_area = self.settings["workspaces_lustre_root_dir"]
        wf_dir = self.settings["destination_subdir"]
        return Path(spool_area + "/" + wf_dir)

    def check_spool_contents(self, spool_path: Path, expected_dirs: List[str]) -> bool:
        """
        Make sure request's spool dir has everything we'd expect for a successful retrieval.

        :param spool_path:
        :param expected_dirs:
        :return:
        """
        os.chdir(spool_path)

        contents = glob.glob("*/", recursive=True)
        if set(contents).issubset(expected_dirs) and not len(contents) < len(expected_dirs):
            self.logger.info("All directories were successfully transferred!")
            return True

        self.logger.error(
            f"ERROR: A directory seems to be missing! "
            f"Directory list {contents} does not match the "
            f"list of expected directories!"
        )
        return False

    def determine_capability_request_and_version_ids(self) -> Tuple[int, int]:
        # We must use our workflow request ID to get the capability execution ID from the capability service
        url = f"{self.settings['capability_url']}/capability/request/execution/{self.settings['request_id']}"
        execution = requests.get(url).json()
        return execution["capability_request_id"], execution["version_number"]

    def convey(self):
        self.logger.info(f"RUNNING POST QA STANDARD {self.action.upper()}!")
        self.retrieval()
